//=============================================================================
//
//  CLASS Torus
//
//=============================================================================
#ifndef RAYTRACER_TORUS_H
#define RAYTRACER_TORUS_H

//== INCLUDES =================================================================
#include "Surface.h"


//== CLASS DEFINITION =========================================================

/*!
	A sphere. The typical first raytracing-surface!
*/
class Torus : public Surface
{

    //-------------------------------------------------------------- private data
private:
    VectorType3 m_v3Position;
    double    m_c;
	double    m_a;


public:

    //-------------------------------------------------------------- constructors

    //! standard constructor
    Torus()
    {
    };

    //! initializing constructor
    Torus(
        const VectorType3   &v3Position,  
        const double      &c,
		const double      &a )  :
    m_v3Position( v3Position ),
    m_c( c ),
	m_a( a )
    {
    };

    //-------------------------------------------------------------- element access
    void setPosition(
        const VectorType3   &v3Position ) 
    {
        m_v3Position = v3Position;
    };

    
    const VectorType3 &getPosition() const
    {
        return m_v3Position;
    };

    
    void setA(
        const double  &a )  
    {
        m_a = a;
    };

    
    const double getA() const
    {
        return m_a;
    };

    
    void setC(
        const double  &c )  
    {
        m_c = c;
    };

    
    const double getC() const
    {
        return m_c;
    };

    
    AABBox getAABB() const
    {
        VectorType3 v3_offset = VectorType3( m_a + m_c, m_c, m_a + m_c );
        return AABBox( m_v3Position - v3_offset, m_v3Position + v3_offset );
    };

    //render object in OpenGL
    void    renderGL() const;

    //translate object
    void    translate( float x, float y, float z );

    //rotate object
    void    rotate( float rotx, float roty, float rotz );

    //get Translation
    void getTranslation(
        float   *x,     
        float   *y,     
        float   *z )    
    {
        *x = m_v3Position[0];
        *y = m_v3Position[1];
        *z = m_v3Position[2];
    };

    //get Rotation
    void getRotation(
        int *rotx,  
        int *roty,  
        int *rotz ) 
    {
        *rotx = 0;
        *roty = 0;
        *rotz = 0;
    };

    //--------------------------------------------------------------- intersections
    bool    intersect( const Ray       &clRay,
                       const double  t0,
                       const double  t1,
                       TracingContext  &tContext ) const;
};

//=============================================================================
#endif // RAYTRACER_TORUS_H

