/** 
*
* @author
*   Gabriel Zachmann, Sep 2006
*
* @bug
*   I'm sure there is still some bugs left ....
**/
#include <QtGui>
#include <QtOpenGL>
#include <QMouseEvent>
#include <math.h>
#include "glwidget.h"

GLWidget::GLWidget( QWidget * father ) : QGLWidget( father ), m_xRot( 0.0 ), m_yRot( 0.0 ), m_zRot( 0.0 ), m_xTrans( 0.0 ), m_yTrans( 0.0 ), m_zTrans( -5.0 ), m_scene( NULL ), m_selectedSurface( NULL ), m_raytracer( NULL ), m_glRays( NULL ), m_showRays( false )
{
    setFocusPolicy( Qt::ClickFocus );
}


GLWidget::~GLWidget()
{
    makeCurrent();
}


QSize GLWidget::minimumSizeHint() const
{
    return QSize( 100, 100 );
}


QSize GLWidget::sizeHint() const
{
    return QSize( 400, 400 );
}

/* Set some OpenGL states */
void GLWidget::initializeGL()
{

    // set opengl states
    glEnable( GL_DEPTH_TEST );
    glEnable( GL_CULL_FACE );
    glEnable( GL_NORMALIZE );
}

// Warning: don't make paintGL() and resizeGL() const!
// otherwise, the rendering will produce complete garbage.

// (is this a bug in Qt? g++? opengl?)
void GLWidget::paintGL()
{
    if( m_raytracer )
    {
		ColorType background = m_raytracer->getBackgroundColor();
		glClearColor( background.red(), background.green(), background.blue(), 1 );
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		//Enable Lights
		m_raytracer->renderLightsGL();

        //render Scene
        if( m_scene != NULL )
            m_scene->renderGL();
        if( m_glRays != NULL && m_showRays )
		{
            		drawRays();
#ifdef WIN32
            		Sleep( 100 ); //Wait a little bit
#else

			usleep( 100 * 1000 );
#endif
		}
    }
}


void GLWidget::resizeGL(
    int w,  
    int h ) // = width & height
{
    int side = qMin( w, h );
    glViewport( (w - side) / 2, (h - side) / 2, w, h );
}

/* Process Keyboard Events */
void GLWidget::keyPressEvent(
    QKeyEvent   *e )    
{
    if( !isValid() )
        return;
    switch( e->key() )
    {
        case Qt::Key_B: break;
        default:        return;
    }
}


/* Change Rotation */
void GLWidget::setXRotation(
    int angle ) 
{
    if( m_selectedSurface != NULL )
    {
        m_selectedSurface->rotate( angle, 0, 0 );
        updateGL();
    }
}

/* Change Rotation */
void GLWidget::setYRotation(
    int angle ) 
{
    if( m_selectedSurface != NULL )
    {
        m_selectedSurface->rotate( 0, angle, 0 );
        updateGL();
    }
}

/* Change Rotation */
void GLWidget::setZRotation(
    int angle ) 
{
    if( m_selectedSurface != NULL )
    {
        m_selectedSurface->rotate( 0, 0, angle );
        updateGL();
    }
}

/* Change Translation */
void GLWidget::setXTranslation(
    int value ) 
{
    if( m_selectedSurface != NULL )
    {
        m_xTrans = m_selectedNodeOrgTranslation[0] +
            ( value / 100.0 ) *
            maxTranslation;
        m_selectedSurface->translate( m_xTrans, m_yTrans, m_zTrans );
        updateGL();
    }
}

/* Change Translation */
void GLWidget::setYTranslation(
    int value ) 
{
    if( m_selectedSurface != NULL )
    {
        m_yTrans = m_selectedNodeOrgTranslation[1] +
            ( value / 100.0 ) *
            maxTranslation;
        m_selectedSurface->translate( m_xTrans, m_yTrans, m_zTrans );
        updateGL();
    }
}

/* Change Translation */
void GLWidget::setZTranslation(
    int value ) 
{
    if( m_selectedSurface != NULL )
    {
        m_zTrans = m_selectedNodeOrgTranslation[2] +
            ( value / 100.0 ) *
            maxTranslation;
        m_selectedSurface->translate( m_xTrans, m_yTrans, m_zTrans );
        updateGL();
    }
}

/* Set Active Object */
void GLWidget::selectNextObject()
{
    if( m_raytracer != NULL )
    {
        m_selectedSurface = m_scene->selectNext();

        int     xRot, yRot, zRot;
        float   xTrans, yTrans, zTrans;
        m_selectedSurface->getRotation( &xRot, &yRot, &zRot );
        m_selectedSurface->getTranslation( &xTrans, &yTrans, &zTrans );
        m_selectedNodeOrgTranslation[0] = xTrans;
        m_selectedNodeOrgTranslation[1] = yTrans;
        m_selectedNodeOrgTranslation[2] = zTrans;
        m_xTrans = xTrans;
        m_yTrans = yTrans;
        m_zTrans = zTrans;

        //set Sliders
        emit    xRotationChanged( xRot );
        emit    yRotationChanged( yRot );
        emit    zRotationChanged( zRot );
        emit    xTranslationChanged( 0 );
        emit    yTranslationChanged( 0 );
        emit    zTranslationChanged( 0 );
        updateGL();
    }
}

/* Wether or not Ray are rendered */
void GLWidget::setRayMode(
    int mode )  
{
	if( mode == Qt::Unchecked )
		m_showRays = false;
	if( mode == Qt::Checked )
		m_showRays = true;
}

/* Set Geometry, Lights, Camera etc */
void GLWidget::setScene(
    Raytracer *raytracer, 
    SurfaceList      *scene )    
{
    m_raytracer = ( Raytracer * ) raytracer;
    m_scene = ( SurfaceList * ) scene;

    VectorType3  maxT;
    maxT = m_raytracer->getCamera().getEye() - m_raytracer->getCamera().getLookAt();
    maxTranslation = maxT.length();
    glViewport( 0, 0, width(), height() );
    glMatrixMode( GL_PROJECTION );  // Kameraparameter!
    glLoadIdentity();   // zuruecksetzen
    double left, right, bottom, top;
    GLfloat             nearplane = m_raytracer->getCamera().getFocalLength();
    GLfloat             farplane = 100.0 * nearplane;
    m_raytracer->getCamera().getWindow( &left, &bottom, &right, &top );
    glFrustum( left, right, bottom, top, nearplane, farplane );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    //setup Camera
    gluLookAt( m_raytracer->getCamera().getEye()[0],
               m_raytracer->getCamera().getEye()[1],
               m_raytracer->getCamera().getEye()[2],
               m_raytracer->getCamera().getLookAt()[0],
               m_raytracer->getCamera().getLookAt()[1],
               m_raytracer->getCamera().getLookAt()[2],
               m_raytracer->getCamera().getUp()[0],
               m_raytracer->getCamera().getUp()[1],
               m_raytracer->getCamera().getUp()[2] );
    updateGL();
}

/* Change Ray-Rendering status */
void GLWidget::renderRays(
    RayArray *glRays )   
{
    if( m_showRays )
    {
        m_glRays = glRays;
        updateGL();
    }
}

/* Stop Ray rendering when finished raytracing */
void GLWidget::reset()
{
	if ( m_raytracer != NULL )
	{
		m_glRays = NULL;	
		VectorType3  maxT;
		maxT = m_raytracer->getCamera().getEye() - m_raytracer->getCamera().getLookAt();
		maxTranslation = maxT.length();
		glViewport( 0, 0, width(), height() );
		glMatrixMode( GL_PROJECTION );  // Kameraparameter!
		glLoadIdentity();   // zuruecksetzen
		double left, right, bottom, top;
		GLfloat             nearplane = m_raytracer->getCamera().getFocalLength();
		GLfloat             farplane = 100.0 * nearplane;
		m_raytracer->getCamera().getWindow( &left, &bottom, &right, &top );
		glFrustum( left, right, bottom, top, nearplane, farplane ); 
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		//setup Camera
		gluLookAt( m_raytracer->getCamera().getEye()[0],
				   m_raytracer->getCamera().getEye()[1],
				   m_raytracer->getCamera().getEye()[2],
				   m_raytracer->getCamera().getLookAt()[0],
				   m_raytracer->getCamera().getLookAt()[1],
				   m_raytracer->getCamera().getLookAt()[2],
				   m_raytracer->getCamera().getUp()[0],
				   m_raytracer->getCamera().getUp()[1],
				   m_raytracer->getCamera().getUp()[2] );
		updateGL();
	}
}

/* Draw Rays if m_showRays is true */
void GLWidget::drawRays()
{
    glPushMatrix();
    glDisable( GL_DEPTH_TEST );
	glDisable( GL_LIGHTING );
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glColor4f( 1, 1, 1, 0.3);

	m_imag = QImage( *m_textureImage );
	bindTexture( QPixmap::fromImage( m_imag ), GL_TEXTURE_2D);
	
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND );
	glEnable(GL_TEXTURE_2D);

	glBegin( GL_QUADS );
	double left, right, bottom, top;
	m_raytracer->getCamera().getWindow( &left, &bottom, &right, &top );
	GLfloat  nearplane = m_raytracer->getCamera().getFocalLength();

	glTexCoord2f( 0.0, 0.0 );
	glVertex3f( left, bottom, nearplane );
	glTexCoord2f( 1.0, 0.0 );
	glVertex3f( right, bottom, nearplane );
	glTexCoord2f( 1.0, 1.0 );
	glVertex3f( right, top, nearplane );
	glTexCoord2f( 0.0, 1.0 );
	glVertex3f( left, top, nearplane );
	glEnd();

	if ( m_imag.width() < 25 )
	{
		glBegin( GL_LINES );
		for ( int i = 0; i < m_imag.width(); i++ )
		{
			glVertex3f( left+i*(right-left)/(float) m_imag.width(), top, nearplane );
			glVertex3f( left+i*(right-left)/(float) m_imag.width(), bottom, nearplane );
		}

		for ( int i = 0; i < m_imag.height(); i++ )
		{
			glVertex3f( left, bottom+i*(top-bottom)/(float) m_imag.height(), nearplane );
			glVertex3f( right, bottom+i*(top-bottom)/(float) m_imag.height(), nearplane );
		}
		glEnd();
	}
	
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);

	glBegin( GL_LINES );
	glColor4f( 1, 1, 1, 1);
    for( unsigned int i = 0; i < m_glRays->size(); i++ )
    {
        glVertex3f( (*m_glRays)[i].m_start[0], (*m_glRays)[i].m_start[1],
                    (*m_glRays)[i].m_start[2] );
        glVertex3f( (*m_glRays)[i].m_stop[0], (*m_glRays)[i].m_stop[1],
                    (*m_glRays)[i].m_stop[2] );
    }

    glEnd();
    glEnable( GL_DEPTH_TEST );
	glEnable( GL_LIGHTING );
    glPopMatrix();
}

void GLWidget::startRaytracing( QImage *textureImage )
{
	if ( m_raytracer != NULL )
	{
		m_textureImage = textureImage;

		VectorType3  maxT;
		maxT = m_raytracer->getCamera().getEye() - m_raytracer->getCamera().getLookAt();
		maxTranslation = maxT.length();
		glViewport( 0, 0, width(), height() );
		glMatrixMode( GL_PROJECTION );  // Kameraparameter!
		glLoadIdentity();   // zuruecksetzen
		double left, right, bottom, top;
		GLfloat             nearplane = m_raytracer->getCamera().getFocalLength();
		GLfloat             farplane = 100.0 * nearplane;
		m_raytracer->getCamera().getWindow( &left, &bottom, &right, &top );
		glOrtho(3*left, 3*right, 3*bottom, 3*top, nearplane, farplane); 
		glMatrixMode( GL_MODELVIEW );
		
        //setup Camera
        gluLookAt( 2*m_raytracer->getCamera().getEye()[0],
                   2*m_raytracer->getCamera().getEye()[1],
                   2*m_raytracer->getCamera().getEye()[2],
                   m_raytracer->getCamera().getLookAt()[0],
                   m_raytracer->getCamera().getLookAt()[1],
                   m_raytracer->getCamera().getLookAt()[2],
                   m_raytracer->getCamera().getUp()[0],
                   m_raytracer->getCamera().getUp()[1],
                   m_raytracer->getCamera().getUp()[2] );
	}
}